/*
 * Bufwrite:  Take output from a program and buffer all of its output
 *            into a single write to stdout.  Used for "weak" atomic
 *            writes for programs that have small amounts of output,
 *            like traceroute.  No guarantees about the atomicity
 *            of the output, but it works well enough.
 *
 * Author:  David G. Andersen
 * June 2000.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/param.h>

void
bufwrite(char *program_path, char *progname,
	 char *remhost, char *progargs, int oneline)
{
    FILE *progout;
    char progargpath[512];
    char buf[128];
    char *tbuf = NULL;
    int curbufsize = 1024;
    int curbufcontents = 0;
    int justread;

    tbuf = malloc(curbufsize);
    if (!tbuf) {
	perror("could not allocate buffer for traceroute");
	exit(-1);
    }

    curbufcontents = sprintf(tbuf, "||%s|%s|%ld||%s",
			     progname, remhost, time(NULL),
			     oneline ? " " : "\n");

    snprintf(progargpath, sizeof(progargpath)-1, "%s %s %s 2>&1",
	     program_path, progargs, remhost);

    progout = popen(progargpath, "r");
    if (progout == NULL) {
	perror(program_path);
	exit(-1);
    }
    while (fgets(buf, sizeof(buf)-1, progout)) {
	justread = strlen(buf);
	while ((justread + curbufcontents + 2)> curbufsize) {
	    curbufsize *= 2;
	    tbuf = realloc(tbuf, curbufsize);
	    if (tbuf == NULL) {
		perror("could not resize bufwrite buffer");
		exit(-1);
	    }
	}
	/* Now we're the proper size */
	memcpy(tbuf+curbufcontents, buf, justread);
	curbufcontents += justread;
	tbuf[curbufcontents] = 0;
    }
    write(STDOUT_FILENO, tbuf, curbufcontents);
    return;
}

void
usage()
{
    fprintf(stderr,
	    "bufwrite <program path> <prog name> <remote host> <prog args>\n"
	    );
}

void
help()
{
    usage();
    fprintf(stderr,
	    " -h ... help (this message)\n"
	    " -o ... don't add newline to the log message\n"
	    " argv[1] should be the path of the program to execute\n"
	    " argv[2] should be the name of the program to appear in the log\n"
	    " argv[3] should be the remote host to contact\n"
	    " argv[4] should be extra args to the program\n"
	    );
}

int
main(int argc, char **argv)
{
    extern char *optarg;
    extern int optind;
    int ch;
    int oneline = 0;

    while ((ch = getopt(argc, argv, "ho")) != -1)
	 switch (ch) {
	 case 'o':
	     oneline = 1;
	     break;
	 case 'h':
	     help();
	     exit(0);
	 default:
	     usage();
	     exit(-1);
	 }
     argc -= optind;
     argv += optind;

     if (argc != 4) {
	 usage();
	 exit(-1);
     }

     if (strlen(argv[0]) > MAXPATHLEN) {
	 fprintf(stderr, "program path %s too long\n", argv[0]);
	 exit(-1);
     }

     if (strlen(argv[2]) > MAXHOSTNAMELEN) {
	 fprintf(stderr, "host name %s too long\n", argv[2]);
	 exit(-1);
     }

     bufwrite(argv[0], argv[1], argv[2], argv[3], oneline);

     exit(0);
}
